import requests
import logging
import typing as t


class CursorApiIterator:
    def __init__(self,
                 session: requests.Session,
                 url: str,
                 method: str,
                 response_cursor_field_name: str,
                 response_records_field_name: str,
                 request_cursor_param_name: t.Optional[str] = None,
                 request_cursor_field_name: t.Optional[str] = None,
                 extra_params: t.Optional[dict] = None,
                 extra_data: t.Optional[dict] = None,
                 extra_headers: t.Optional[dict] = None,
                 empty_data_on_next_cursor: bool=False,
                 payload_handler=None):
        """

        :param session: requests.Session object (you can set session.auth = (user, passwd) for authentication)
        :param url: url to call
        :param method: get or post
        :param response_cursor_field_name:
        :param response_records_field_name:
        :param request_cursor_param_name:
        :param request_cursor_field_name:
        :param extra_params: URL params
        :param extra_data: request payload params
        :param extra_headers:
        :param empty_data_on_next_cursor: Flase
        :param payload_handler: a function that is called just after fetching next page. Takes dict as a param.
        """
        self.logger = logging.getLogger(__name__)
        self._session = session
        self._url = url
        self._method = method.upper()
        self._request_cursor_param_name = request_cursor_param_name
        self._request_cursor_field_name = request_cursor_field_name
        self._response_cursor_field_name = response_cursor_field_name
        self._response_records_field_name = response_records_field_name
        self._extra_params = extra_params
        self._extra_data = extra_data
        self._request_headers = {
            "Accept": "application/json",
        }
        if self._request_cursor_field_name:
            self._request_headers["Content-type"] = "application/json"
        if extra_headers:
            for k, v in extra_headers.items():
                self._request_headers[k] = v
        self._empty_data_on_next_cursor = empty_data_on_next_cursor
        self._payload_handler = payload_handler
        self._completed = False
        self._next_cursor = None
        self.page_number = 1

    def __iter__(self):
        self.logger.debug(f"Downloading from {self._url}")
        self._completed = False
        self._next_cursor = None
        self.page_number = 1
        return self

    def __next__(self):
        if self._completed:
            raise StopIteration
        if self._extra_params or self._request_cursor_param_name:
            params = {}
            if self._extra_params:
                params = {**self._extra_params}
            params[self._request_cursor_param_name] = self._next_cursor
        else:
            params = None

        if self._extra_data or self._request_cursor_field_name:
            data = {}

            if self._extra_data and (not self._empty_data_on_next_cursor or not self._next_cursor):
                data = {**self._extra_data}

            if self._request_cursor_field_name and self._next_cursor:
                data[self._request_cursor_field_name] = self._next_cursor
        else:
            data = None

        response = self._session.request(method=self._method, url=self._url, headers=self._request_headers, params=params, json=data)
        response.raise_for_status()

        response_json = response.json()

        if self._payload_handler:
            self._payload_handler(response_json)

        if self._response_records_field_name not in response_json.keys():
            raise Exception(f"Failed to get field '{self._response_records_field_name}' from response {response_json}. Request params: '{params}' data: '{data}'")

        self._next_cursor = response_json.get(self._response_cursor_field_name)
        self._completed = not self._next_cursor
        self.logger.debug(
            f"Downloaded page {self.page_number}. Items collected: {len(response_json[self._response_records_field_name])}")
        self.page_number += 1

        return response_json[self._response_records_field_name]





class NormalizeException(Exception):
    def __init__(self, path, message):
        super().__init__(f"{path}: {message}")
        self._path = path
        self._message = message


def normalize_with_prototype(prototype, object_to_norm, freestyle_attrs_name="freestyle_attrs", pass_through_paths=[],
                             strict_types=True, type_converter=None):
    """
    Normalize python dict, so that it has all the fields and only the fields specified in a prototype dict.

    :param prototype: A prototype dict, for example from Swagger doc.
    :param object_to_norm: A dict that comes from API.
    :param freestyle_attrs_name: it is a name for attribute, when prototype contains a dict like this: "field": {"": ""}
    :param pass_through_paths: list of paths which will be passed through without normalization (for example dicts with
        unpredicted structure, that will be handled after normalization)
    :param strict_types: if True, then fail on types mismatch between prototype and object_to_norm.
        If False, function will attempt to parse string, i.e. "123.45" to 123.45 float.
        It will fail on parsing errors anyway.
    :param type_converter: a function with signature:
        type_converter(prototype_object: any, object_to_norm: any, path_info: str) -> any
        called only when strict_types=False and type(prototype_object) != type(object_to_norm)
    :return:
    """
    if not type_converter:
        type_converter = lambda prototype_object, object_to_norm, path_info: object_to_norm

    def normalize_with_prototype_rec(prototype, object_to_norm, path_info, prototype_path):
        try:
            if prototype_path in pass_through_paths:
                return object_to_norm
            if isinstance(prototype, dict):
                if object_to_norm is None:
                    object_to_norm = {}
                elif not isinstance(object_to_norm, dict) and not strict_types:
                    object_to_norm = type_converter(prototype, object_to_norm, path_info)
                if not isinstance(object_to_norm, dict):
                    raise NormalizeException(path_info, f"object_to_norm is not dict (of value {object_to_norm})")
                normalized_object = {}
                for k, v in prototype.items():
                    if k != "":
                        normalized_object[k] = normalize_with_prototype_rec(prototype[k], object_to_norm.get(k), path_info=f"{path_info}.{k}", prototype_path=f"{prototype_path}.{k}" if prototype_path != "" else k)
                    else:
                        normalized_object[freestyle_attrs_name] = [
                            {"key": key, "value": value} for key, value in object_to_norm.items() if key not in prototype.keys() or key == ""
                        ]
                return normalized_object
            elif isinstance(prototype, list):
                if object_to_norm is None:
                    object_to_norm = []
                elif not isinstance(object_to_norm, list) and not strict_types:
                    object_to_norm = type_converter(prototype, object_to_norm, path_info)
                if not isinstance(object_to_norm, list):
                    raise NormalizeException(path_info, f"object_to_norm is not list (of value {object_to_norm})")
                return [normalize_with_prototype_rec(prototype[0], e, path_info=f"{path_info}[{idx}]", prototype_path=prototype_path) for idx, e in enumerate(object_to_norm)]
            elif isinstance(prototype, str):
                if object_to_norm is None:
                    return None
                if type(prototype) != type(object_to_norm):
                    if strict_types:
                        raise NormalizeException(path_info, f"type(prototype) != type(object_to_norm): {type(prototype)} != {type(object_to_norm)} (of value {object_to_norm})")
                    else:
                        return str(type_converter(prototype, object_to_norm, path_info))
                return object_to_norm
            elif isinstance(prototype, bool):
                if object_to_norm is None:
                    return None
                if type(prototype) != type(object_to_norm):
                    if strict_types:
                        raise NormalizeException(path_info, f"type(prototype) != type(object_to_norm): {type(prototype)} != {type(object_to_norm)} (of value {object_to_norm})")
                    else:
                        object_to_norm = type_converter(prototype, object_to_norm, path_info)
                        if not isinstance(object_to_norm, bool):
                            if isinstance(object_to_norm, str):
                                if object_to_norm.lower().strip() in {'1', 'true', 'yes'}:
                                    object_to_norm = True
                                elif object_to_norm.lower().strip() in {'0', 'false', 'no'}:
                                    object_to_norm = False
                                else:
                                    object_to_norm = None
                            else:
                                object_to_norm = bool(object_to_norm)
                return object_to_norm
            elif isinstance(prototype, int):
                if object_to_norm is None:
                    return None
                if not isinstance(object_to_norm, int):
                    if strict_types:
                        raise NormalizeException(path_info, f"type(prototype) != type(object_to_norm): {type(prototype)} != {type(object_to_norm)} (of value {object_to_norm})")
                    else:
                        if not object_to_norm or isinstance(object_to_norm, str) and object_to_norm.strip() == "":
                            return None
                        try:
                            object_to_norm = int(type_converter(prototype, object_to_norm, path_info))
                        except Exception as e:
                            raise NormalizeException(path_info, f"can't parse int '{object_to_norm}': {e}")
                return object_to_norm
            elif isinstance(prototype, float):
                if object_to_norm is None:
                    return None
                if not isinstance(object_to_norm, (int, float)):
                    if strict_types:
                        raise NormalizeException(path_info, f"type(prototype) != type(object_to_norm): {type(prototype)} != {type(object_to_norm)} (of value {object_to_norm})")
                    else:
                        if not object_to_norm or isinstance(object_to_norm, str) and object_to_norm.strip() == "":
                            return None
                        try:
                            object_to_norm = float(type_converter(prototype, object_to_norm, path_info))
                        except Exception as e:
                            raise NormalizeException(path_info, f"can't parse float '{object_to_norm}': {e}")
                if not isinstance(object_to_norm, float):
                    try:
                        object_to_norm = float(object_to_norm)
                    except Exception as e:
                        raise NormalizeException(path_info, f"can't parse float '{object_to_norm}': {e}")
                return object_to_norm
            else:
                raise NormalizeException(path_info, f"prototype data type not supported: {type(prototype)}")
        except NormalizeException as e:
            raise
        except Exception as e:
            NormalizeException(path_info, str(e))
    return normalize_with_prototype_rec(prototype, object_to_norm, "", "")



from pprint import pprint
from bezalel import normalize_dicts
from unittest import TestCase
import json


def test_simple_case_1():
    data = [
        {
            "id": 1, "name": "John Smith",
            "pets": [
                {"id": 101, "type": "cat", "name": "Kitty", "toys": [{"name": "toy1"}, {"name": "toy2"}]},
                {"id": 102, "type": "dog", "name": "Barky", "toys": [{"name": "toy3"}]}
            ]
        },
        {
            "id": 2, "name": "Sue Smith",
            "pets": [
                {"id": 201, "type": "cat", "name": "Kitten", "toys": [{"name": "toy4"}, {"name": "toy5"}, {"name": "toy6"}]},
                {"id": 202, "type": "dog", "name": "Fury", "toys": []}
            ]
        },
    ]

    expected_list = [
        {'id': 1, 'name': 'John Smith', 'pets.id': 101, 'pets.type': 'cat', 'pets.name': 'Kitty', 'pets.toys.name': 'toy1'},
        {'id': 1, 'name': 'John Smith', 'pets.id': 101, 'pets.type': 'cat', 'pets.name': 'Kitty', 'pets.toys.name': 'toy2'},
        {'id': 1, 'name': 'John Smith', 'pets.id': 102, 'pets.type': 'dog', 'pets.name': 'Barky', 'pets.toys.name': 'toy3'},
        {'id': 2, 'name': 'Sue Smith', 'pets.id': 201, 'pets.type': 'cat', 'pets.name': 'Kitten', 'pets.toys.name': 'toy4'},
        {'id': 2, 'name': 'Sue Smith', 'pets.id': 201, 'pets.type': 'cat', 'pets.name': 'Kitten', 'pets.toys.name': 'toy5'},
        {'id': 2, 'name': 'Sue Smith', 'pets.id': 201, 'pets.type': 'cat', 'pets.name': 'Kitten', 'pets.toys.name': 'toy6'},
        {'id': 2, 'name': 'Sue Smith', 'pets.id': 202, 'pets.type': 'dog', 'pets.name': 'Fury'}]
    result = normalize_dicts(data, ["pets", "toys"])
    TestCase().assertListEqual(expected_list, result)


def test_empty_input():
    data = []
    expected_list = []
    result = normalize_dicts(data, ["pets", "toys"])
    TestCase().assertListEqual(expected_list, result)


def test_with_empty_list():
    data = [{
        "id": 1, "name": "Cole Volk", "fitness": [{"height": 130, "weight": 60}, {"height": 1300, "weight": 600}],
        "someList": []
    },
    {
        "id": 2, "name": "Faye Raker", "fitness": [{"height": 130, "weight": 60}, {"height": 888, "weight": 777}],
        "someList": [1]
    },
    {
        "id": 2, "name": "Faye Raker", "fitness": [{"height": 130, "weight": 60}],
        "someList": [1]
    },
    ]
    expected_list = [
        {'id': 1, 'name': 'Cole Volk', 'someList': [], 'fitness.height': 130, 'fitness.weight': 60},
        {'id': 1, 'name': 'Cole Volk', 'someList': [], 'fitness.height': 1300, 'fitness.weight': 600},
        {'id': 2, 'name': 'Faye Raker', 'someList': [1], 'fitness.height': 130, 'fitness.weight': 60},
        {'id': 2, 'name': 'Faye Raker', 'someList': [1], 'fitness.height': 888, 'fitness.weight': 777},
        {'id': 2, 'name': 'Faye Raker', 'someList': [1], 'fitness.height': 130, 'fitness.weight': 60}]
    result = normalize_dicts(data, ["fitness"])
    print(result)
    TestCase().assertListEqual(expected_list, result)


def test_empty_lists_2():
    data = [{
        "id": 1, "name": "Cole Volk", "fitness": [{"height": 130, "weight": 60}],
        "someList": []
    },
    {
        "id": 2, "name": "Faye Raker", "fitness": [{"height": 130, "weight": 60}],
        "someList": []
    }, ]
    expected_list = [
        {'id': 1, 'name': 'Cole Volk', 'someList': [], 'fitness.height': 130, 'fitness.weight': 60},
        {'id': 2, 'name': 'Faye Raker', 'someList': [], 'fitness.height': 130, 'fitness.weight': 60}]
    result = normalize_dicts(data, ["fitness"])
    print(result)
    TestCase().assertListEqual(expected_list, result)


def test_2_level():
    data = [{
        "id": 1, "name": "Cole Volk", "fitness": [
            {"fit1": 101, "fit2": [{"height": 1001, "weight": 1002}, {"height": 1011, "weight": 1012}, {"height": 1022, "weight": 1023}] },
            {"fit1": 102, "fit2": [{"height": 11, "weight": 22}]},
        ],
        "someList": []
    },
    {
        "id": 2, "name": "Faye Raker", "fitness": [
            {"fit1": 201,
             "fit2": [{"height": 131, "weight": 601}, {"height": 132, "weight": 602}, {"height": 133, "weight": 603}]},
            {"fit1": 202, "fit2": [{"height": 11, "weight": 22}]},
        ],
        "someList": []
    },
    {},
    {"id": 3, "name": "AAA BBB", "fitness": []},
    {"id": 4, "name": "AAA BBB", "fitness": None},
    # {"id": 4, "name": "AAA BBB", "fitness": "not-a-list"} # TODO: it should throw an exception, so move it to another test
    ]

    expected_list = [{'id': 1, 'name': 'Cole Volk', 'someList': [], 'fitness.fit1': 101, 'fitness.fit2.height': 1001, 'fitness.fit2.weight': 1002},
                     {'id': 1, 'name': 'Cole Volk', 'someList': [], 'fitness.fit1': 101, 'fitness.fit2.height': 1011, 'fitness.fit2.weight': 1012},
                     {'id': 1, 'name': 'Cole Volk', 'someList': [], 'fitness.fit1': 101, 'fitness.fit2.height': 1022, 'fitness.fit2.weight': 1023},
                     {'id': 1, 'name': 'Cole Volk', 'someList': [], 'fitness.fit1': 102, 'fitness.fit2.height': 11, 'fitness.fit2.weight': 22},
                     {'id': 2, 'name': 'Faye Raker', 'someList': [], 'fitness.fit1': 201, 'fitness.fit2.height': 131, 'fitness.fit2.weight': 601},
                     {'id': 2, 'name': 'Faye Raker', 'someList': [], 'fitness.fit1': 201, 'fitness.fit2.height': 132, 'fitness.fit2.weight': 602},
                     {'id': 2, 'name': 'Faye Raker', 'someList': [], 'fitness.fit1': 201, 'fitness.fit2.height': 133, 'fitness.fit2.weight': 603},
                     {'id': 2, 'name': 'Faye Raker', 'someList': [], 'fitness.fit1': 202, 'fitness.fit2.height': 11, 'fitness.fit2.weight': 22},
                     {},
                     {'id': 3, 'name': 'AAA BBB'},
                     {'id': 4, 'name': 'AAA BBB'}
                     ]
    result = normalize_dicts(data, ["fitness", "fit2"])
    print(result)
    TestCase().assertListEqual(expected_list, result)


def test_flatten_dict():
    data = [{
        "id": 1, "name": "Cole Volk", "fitness": [{"height": 130, "weight": 60}],
        "someDict": {"a": 101, "b": 102, "c": {"k1": 103, "k2": 104}}
    },
    {
        "id": 2, "name": "Faye Raker", "fitness": [{"height": 130, "weight": 60}],
        "someDict": {"a": 201, "b": 202}, "someList": [1,2,3]
    }, ]
    expected_list = [
        {'id': 1, 'name': 'Cole Volk', "someDict.a": 101, "someDict.b": 102, "someDict.c.k1": 103, "someDict.c.k2": 104, 'fitness.height': 130, 'fitness.weight': 60},
        {'id': 2, 'name': 'Faye Raker', "someDict.a": 201, "someDict.b": 202, "someList": json.dumps([1,2,3]), 'fitness.height': 130, 'fitness.weight': 60}]
    result = normalize_dicts(data, ["fitness"], jsonify_lists=True)
    print(result)
    TestCase().assertListEqual(expected_list, result)


def test_empty_path():
    data = [{
        "id": 1, "name": "Cole Volk", "fitness": [{"height": 130, "weight": 60}],
        "someDict": {"a": 101, "b": 102, "c": {"k1": 103, "k2": 104}}
    },
    {
        "id": 2, "name": "Faye Raker", "fitness": [{"height": 130, "weight": 60}],
        "someDict": {"a": 201, "b": 202}, "someList": [1,2,3]
    }, ]
    expected_list = [
        {'id': 1, 'name': 'Cole Volk', 'fitness': json.dumps([{"height": 130, "weight": 60}]), "someDict.a": 101, "someDict.b": 102, "someDict.c.k1": 103, "someDict.c.k2": 104},
        {'id': 2, 'name': 'Faye Raker', 'fitness': json.dumps([{"height": 130, "weight": 60}]), "someDict.a": 201, "someDict.b": 202, "someList": json.dumps([1,2,3])}
    ]
    result = normalize_dicts(data, [], jsonify_lists=True)
    print(result)
    TestCase().assertListEqual(expected_list, result)


def test_single_record_with_no_elems():
    # this shows that without a prototype, we can't rebuild exact columns in output dictionary
    data = [
    {
        "id": 2, "name": "Faye Raker", "fitness": [],
        "someList": [1]
    },
    ]
    expected_list = [
        {'id': 2, 'name': 'Faye Raker', 'someList': [1]}    # not sure if this makes sense, but this is how it works.
        # {'id': 2, 'name': 'Faye Raker', 'someList': [1], 'fitness.height': None, 'fitness.weight': None}  # current API doesn't allow this
    ]
    result = normalize_dicts(data, ["fitness"])
    print(result)
    TestCase().assertListEqual(expected_list, result)


def test_jsonify_dicts():
    data = [{
        "id": 1, "name": "Cole Volk", "fitness": [{"height": 130, "weight": 60}],
        "someDict": {"a": 101, "b": 102, "c": {"k1": 103, "k2": 104}}
    },
    {
        "id": 2, "name": "Faye Raker", "fitness": [{"height": 130, "weight": 60}],
        "someDict": {"a": 201, "b": 202}, "someList": [1, 2, 3]
    }, ]
    expected_list = [
        {'id': 1, 'name': 'Cole Volk', "someDict": json.dumps({"a": 101, "b": 102, "c": {"k1": 103, "k2": 104}}),
         'fitness.height': 130, 'fitness.weight': 60},
        {'id': 2, 'name': 'Faye Raker', "someDict": json.dumps({"a": 201, "b": 202}), "someList": json.dumps([1, 2, 3]),
         'fitness.height': 130, 'fitness.weight': 60}]
    result = normalize_dicts(data, ["fitness"], jsonify_lists=True, jsonify_dicts=["someDict"])
    print(result)
    TestCase().assertListEqual(expected_list, result)


def test_jsonify_dicts_2():
    data = [
        {
            "id": 1, "name": "John Smith",
            "pets": [
                {"id": 101, "type": "cat", "name": "Kitty", "toys": [{"name": "toy1"}, {"name": "toy2"}], "someDict": {"a": 101, "b": 102, "c": {"k1": 103, "k2": 104}}},
                {"id": 102, "type": "dog", "name": "Barky", "toys": [{"name": "toy3"}], "someDict": {"other dict": 1122, "bbb": 202}}
            ]
        },
        {
            "id": 2, "name": "Sue Smith",
            "pets": [
                {"id": 201, "type": "cat", "name": "Kitten", "toys": [{"name": "toy4"}, {"name": "toy5"}, {"name": "toy6"}]},
                {"id": 202, "type": "dog", "name": "Fury", "toys": []}
            ]
        },
    ]

    expected_list = [
        {'id': 1, 'name': 'John Smith', 'pets.id': 101, 'pets.type': 'cat', 'pets.name': 'Kitty', 'pets.toys.name': 'toy1', 'pets.someDict': json.dumps({"a": 101, "b": 102, "c": {"k1": 103, "k2": 104}})},
        {'id': 1, 'name': 'John Smith', 'pets.id': 101, 'pets.type': 'cat', 'pets.name': 'Kitty', 'pets.toys.name': 'toy2', 'pets.someDict': json.dumps({"a": 101, "b": 102, "c": {"k1": 103, "k2": 104}})},
        {'id': 1, 'name': 'John Smith', 'pets.id': 102, 'pets.type': 'dog', 'pets.name': 'Barky', 'pets.toys.name': 'toy3', 'pets.someDict': json.dumps({"other dict": 1122, "bbb": 202})},
        {'id': 2, 'name': 'Sue Smith', 'pets.id': 201, 'pets.type': 'cat', 'pets.name': 'Kitten', 'pets.toys.name': 'toy4'},
        {'id': 2, 'name': 'Sue Smith', 'pets.id': 201, 'pets.type': 'cat', 'pets.name': 'Kitten', 'pets.toys.name': 'toy5'},
        {'id': 2, 'name': 'Sue Smith', 'pets.id': 201, 'pets.type': 'cat', 'pets.name': 'Kitten', 'pets.toys.name': 'toy6'},
        {'id': 2, 'name': 'Sue Smith', 'pets.id': 202, 'pets.type': 'dog', 'pets.name': 'Fury'}]
    result = normalize_dicts(data, ["pets", "toys"], jsonify_dicts=["pets.someDict"])
    tc = TestCase()
    tc.maxDiff = 2000
    tc.assertListEqual(expected_list, result)



def test_jsonify_dicts3():
    data = [{
        "id": 1, "name": "Cole Volk", "fitness": [{"height": 130, "weight": 60, "someDict": {"a": 101, "b": 102, "c": {"k1": 103, "k2": 104}}}],
    },
    {
        "id": 2, "name": "Faye Raker", "fitness": [{"height": 130, "weight": 60,"someDict": {"a": 201, "b": 202}}],
        "someList": [1, 2, 3]
    }, ]
    expected_list = [
        {'id': 1, 'name': 'Cole Volk',
         'fitness.height': 130, 'fitness.weight': 60, "fitness.someDict": json.dumps({"a": 101, "b": 102, "c": {"k1": 103, "k2": 104}}),},
        {'id': 2, 'name': 'Faye Raker',
         'fitness.height': 130, 'fitness.weight': 60, "fitness.someDict": json.dumps({"a": 201, "b": 202}), "someList": json.dumps([1, 2, 3]),}]
    result = normalize_dicts(data, ["fitness"], jsonify_lists=True, jsonify_dicts=["fitness.someDict"])
    print(result)
    TestCase().assertListEqual(expected_list, result)



import logging
import math
import requests

from .dict_utils import *


class PaginatedApiIterator:

    def __init__(self,
                 session: requests.Session,
                 url: str,
                 request_page_number_param_name: str,
                 # response_page_number_field_name: str,
                 response_records_field_name: str,
                 response_page_count_field_name: Optional[str]=None,
                 response_record_count_field_name: Optional[str]=None,
                 records_per_page: Optional[int]=None,
                 extra_params: dict = None,
                 extra_headers: dict = None,
                 extra_data_fields: dict = None,
                 start_page_number_from_1=True,
                 http_method="GET",
                 request_page_number_location: str = "params"
                 ):
        self.logger = logging.getLogger(__name__)
        self._session = session
        self._http_method = http_method
        self._url = url
        self._request_page_number_param_name = request_page_number_param_name
        self._request_page_number_location = request_page_number_location
        # self._response_page_number_field_name = response_page_number_field_name
        if response_page_count_field_name is None and response_record_count_field_name is None or response_page_count_field_name is not None and response_record_count_field_name is not None:
            raise Exception("Exactly one of response_page_count_field_name, response_record_count_field_name must be provided.")
        self._response_page_count_field_name = response_page_count_field_name
        self._response_record_count_field_name = response_record_count_field_name
        if response_record_count_field_name is not None and records_per_page is None:
            raise Exception("If response_record_count_field_name is provided then records_per_page must be also provided.")
        self._records_per_page = records_per_page

        self._response_records_field_name = response_records_field_name
        self._extra_params = extra_params
        if extra_headers:
            self._request_headers = {
                # "Content-type": "application/json",
                "Accept": "application/json",
                **extra_headers
            }
        else:
            self._request_headers = {
                # "Content-type": "application/json",
                "Accept": "application/json"
            }
        self._extra_data_fields = extra_data_fields
        self._start_page_number_from_1 = start_page_number_from_1
        self.page_number = 1 if self._start_page_number_from_1 else 0
        self._completed = False
    def __iter__(self):
        self.logger.debug(f"Downloading from {self._url}")
        self.page_number = 1 if self._start_page_number_from_1 else 0
        self._completed = False
        return self

    def __next__(self):
        if self._completed:
            raise StopIteration

        request_params = None
        if self._extra_params is not None:
            request_params = self._extra_params.copy()

        request_data = None
        if self._extra_data_fields is not None:
            request_data = self._extra_data_fields.copy()

        if self._request_page_number_location == "params":
            if request_params is None:
                request_params = {}
            request_params[self._request_page_number_param_name] = self.page_number
        elif self._request_page_number_location == "data":
            if request_data is None:
                request_data = {}
            dict_set(request_data, self._request_page_number_param_name, self.page_number)
        else:
            raise Exception(f"Wrong parameter value request_page_number_location={self._request_page_number_location}")

        response = self._session.request(method=self._http_method, url=self._url, headers=self._request_headers,
                                         params=request_params, json=request_data)
        response.raise_for_status()

        response_json = response.json()
        if self._response_page_count_field_name is not None:
            page_count = dict_get(response_json, self._response_page_count_field_name)
        else:
            records_count = dict_get(response_json, self._response_record_count_field_name)
            page_count = math.ceil(records_count/self._records_per_page)

        records = dict_get(response_json, self._response_records_field_name, [])

        self.logger.debug(f"Downloaded page {self.page_number} out of {page_count}. Items collected: {len(records)}")

        self.page_number += 1

        if (self._start_page_number_from_1 and self.page_number > page_count) or \
                (not self._start_page_number_from_1 and self.page_number >= page_count):
            self._completed = True

        return records



from typing import Optional


def dict_get(d: dict, path: str, default: any = None, path_split_char: str = ".") -> any:
    elems = path.split(path_split_char)
    q = d
    for e in elems:
        if not isinstance(q, dict):
            return default
        if e in q.keys():
            q = q.get(e)
        else:
            return default
    return q


def dict_set(d: dict, path: str, val: any, path_split_char: str = "."):
    elems = path.split(path_split_char)

    q = d
    for e in elems[:-1]:
        if e in q.keys():
            if isinstance(q.get(e), dict):
                q = q.get(e)
            else:
                q[e] = {}
                q = q.get(e)
        else:
            q[e] = {}
            q = q[e]
    k = elems[-1]
    q[k] = val

